// TODO these imports will not work if @pixi is not a dependency
type PIXIContainer = import("../node_modules/@pixi/display/index").Container;
type PIXIMatrix = import("../node_modules/@pixi/math/index").Matrix;
type PIXITransform = import("../node_modules/@pixi/math/index").Matrix;
type PIXIRenderer = import("../node_modules/@pixi/core/index").Renderer;
type PIXITexture = import("../node_modules/@pixi/core/index").Texture;
type PIXIGraphics = import("../node_modules/@pixi/graphics/index").Graphics;

/**
 * A factory to create a viewer
 */
declare module "@bimdata/2d-engine" {
  export default E2D.makeViewer;
}

declare namespace E2D {
  function makeViewer(cfg: ViewerCfg): Viewer;

  interface ViewerCfg extends LayerOptions {
    canvas: HTMLCanvasElement;
    autoStart?: boolean;
  }

  interface Viewer {
    readonly canvas: HTMLCanvasElement;
    readonly scene: Scene;
    readonly camera: Camera;
    destroy(): void;
    readonly renderer: Renderer;
    readonly textureManager: TextureManager;
    readonly styles: { default: Style; selected: Style; highlight: Style };
    readonly settings: Settings;
    readonly constants: Constants;
    readonly picker: Picker;
    readonly snapper: Snapper;
    readonly ticker: Ticker;
    readonly ui: UI;
    readonly utils: {
      readonly vector2D: Vector2DUtils;
    };
  }

  interface UI extends EventHandler<UIHandlerEvents> {
    connect(el: HTMLElement): void;
    disconnect(): boolean;
  }

  type UIHandlerEvents = {
    click: { canvasPosition: Point; keys: MouseEventKeys };
    "right-click": { canvasPosition: Point; keys: MouseEventKeys };
    move: { canvasPosition: Point; keys: MouseEventKeys };
    drag: { dx: number; dy: number; keys: MouseEventKeys };
    scroll: {
      canvasPosition: Point;
      dx: number;
      dy: number;
      keys: MouseEventKeys;
    };
    exit: undefined;
  };

  enum LINE_CAP {
    BUTT,
    ROUND,
    SQUARE,
  }

  enum LINE_JOIN {
    BEVEL,
    MITER,
    ROUND,
  }

  type Constants = {
    LINE_CAP: {
      BUTT: LINE_CAP;
      ROUND: LINE_CAP;
      SQUARE: LINE_CAP;
    };
    LINE_JOIN: {
      BEVEL: LINE_JOIN;
      MITER: LINE_JOIN;
      ROUND: LINE_JOIN;
    };
    DEFAULT_MODEL_ID: string;
  };

  interface Vector2DUtils {
    distance(v1: Vector2D, v2: Vector2D): number;
    sub(v1: Vector2D, v2: Vector2D): Vector2D;
    add(v1: Vector2D, v2: Vector2D): Vector2D;
    normalize(v: Vector2D): Vector2D;
    length(v: Vector2D): number;
    dot(v1: Vector2D, v2: Vector2D): number;
    cross(v1: Vector2D, v2: Vector2D): number;
    /**
     * @param angle in radians anticlockwise.
     */
    rotateAround(v: Vector2D, center: Vector2D, angle: number): Vector2D;
    /**
     * @returns { number } angle in radians anticlockwise.
     */
    angle(v1: Vector2D, v2: Vector2D): number;
  }

  interface Settings {
    rotateKey: string;
    scaleSpeed: number;
    rotateSpeed: number;
    maxScale: number;
    minScale: number;
    fitViewRatio: number;
    snapMargin: number;
    curves: {
      adaptive: boolean;
      maxLength: number;
      maxSegments: number;
      minSegments: number;
    };
  }

  interface Ticker extends EventHandler<{ tick: number }>, Scheduler {
    start(): void;
    stop(): void;
    running: boolean;
  }

  interface Scheduler {
    tasks: Map<string, Task>;
    add(name: string, fn: Function, priority?: number, once?: boolean): void;
    addOnce(name: string, fn: Function, priority?: number): void;
    remove(name: string): boolean;
    clear(): void;
  }

  interface Task {
    name: string;
    fn: Function;
    priority: number;
    once: boolean;
  }

  interface Picker extends Readonly<EventHandler<PickerEvents>> {
    pick(x: number, y: number): SceneObject | undefined;
  }

  type PickerEvents = {
    pick: {
      object: SceneObject;
      canvasPosition: Point;
      keys: MouseEventKeys;
      rightClick: boolean;
    };
    "pick-nothing": {
      canvasPosition: Point;
      keys: MouseEventKeys;
      rightClick: boolean;
    };
    "hover-surface": { object: SceneObject; canvasPosition: Point };
    hover: { object: SceneObject; canvasPosition: Point };
    "hover-out": { object: SceneObject; canvasPosition: Point };
    "hover-off": { canvasPosition: Point };
  };

  interface Snapper {
    getSnap(canvasPosition: Point, options: SnapOptions): Point;
  }

  interface SnapOptions {
    canvasDistance?: number;
    snapVertices?: boolean;
    snapLines?: boolean;
  }

  interface CameraController {
    destroy(): void;
    translatable: boolean;
    rotatable: boolean;
    scallable: boolean;
  }

  /**
   * A camera allows to see the world. Two coordinate systems can be manipulated using it:
   * - the world coordinates, mentionned as `position`.
   * - the canvas coordinates, mentionned as `canvasPosition`.
   *
   * A camera emits events:
   * - "update", payload { MatrixExtended } : the camera transform. Emited when the camera is updated.
   *
   * To express the zoom level, the camera has two units:
   * - zoom, expressed in percentages, as 100 times the scale.
   * - scale ]0,+Infitiy[.
   *
   */
  interface Camera
    extends Transformable,
      Readonly<EventHandler<{ update: PIXITransform }>> {
    fitView(
      target:
        | E2D.Scene
        | E2D.Scene[]
        | E2D.Model
        | E2D.Model[]
        | E2D.SceneObject
        | E2D.SceneObject[]
    ): void;
    destroy(): void;
    controller: CameraController;
    /**
     * @param factor ]-Infinity, +Infinity[
     * @param origin the zoom transform center.
     */
    zoomIn(factor?: number, origin?: Point): void;
    /**
     * @param factor ]-Infinity, +Infinity[
     * @param origin the zoom transform center.
     */
    zoomOut(factor?: number, origin?: Point): void;
    /**
     * @param angle in degree clockwise.
     * @param origin the rotation transform center.
     */
    transform: PIXIMatrix;
    position: Point;
    /**
     * @returns { number } the camera rotation angle in degree clockwise.
     */
    rotation: number;
    /**
     * @returns { number } ]-Infinity, +Infinity[
     */
    getScale(): number;
    /**
     * @returns { number } ]0, +Infinity[
     */
    zoom: number;
    /**
     * Returns the position of the given canvas position.
     */
    getPosition(canvasPosition: Point): Point;
    /**
     * Returns the canvas position of the given position.
     */
    getCanvasPosition(position: Point): Point;
    getViewpoint(): Viewpoint;
    setViewpoint(viewpoint: Viewpoint): Viewpoint;
  }

  interface Viewpoint {
    x: number;
    y: number;
    /**
     * degree clockwise
     */
    rotation: number;
    /**
     * ]-Infinity, +Infinity[
     */
    zoom: number;
  }

  interface Point {
    x: number;
    y: number;
  }

  interface ReactivePoint extends Point, ReactiveEntity {
    index: number;
  }

  interface Vector2D extends Point {}

  type Events = {
    [eventName: string]: Object;
  };

  interface MouseEventKeys {
    ctrl: boolean;
    alt: boolean;
    shift: boolean;
    meta: boolean;
  }

  interface EventHandler<T extends Events> {
    on<N extends keyof T>(eventName: N, callback: (arg: T[N]) => void): number;
    once<N extends keyof T>(
      eventName: N,
      callback: (arg: T[N]) => void
    ): number;
    off(subscriptionId: number): void;
    emit<N extends keyof T>(eventName: N, payload: T[N]): void;
    clear(): void;
  }

  interface Typed<T> {
    type: T;
  }

  enum EntityType {
    scene = "scene",
    model = "model",
    object = "object",
  }

  interface Scene
    extends Readonly<EventHandler<SceneEvents>>,
      Typed<EntityType.scene> {
    dirty: boolean;
    cache: Object;
    readonly viewer: Viewer;
    generateReport(): SceneReport;
    // Models
    readonly models: Model[];
    readonly modelsMap: Map<string, Model>;
    addModel(modelData: ModelData): Model;
    removeModel(id: string): boolean;
    _removedModels: Model[];
    // Objects
    readonly objects: SceneObject[];
    readonly objectsMap: Map<string, SceneObject>;
    readonly pickingIds: Map<number, SceneObject>;
    addObject(objectData: SceneObjectData, model?: Model): SceneObject;
    removeObject(objectId: string): boolean;
    // Objects setters
    showObjects(ids: string[]): void;
    hideObjects(ids: string[]): void;
    selectObjects(ids: string[]): void;
    deselectObjects(ids: string[]): void;
    highlightObjects(ids: string[]): void;
    unhighlightObjects(ids: string[]): void;
    setObjectsPickable(ids: string[]): void;
    setObjectsUnpickable(ids: string[]): void;
    // Object getters
    readonly shownObjects: SceneObject[];
    readonly hiddenObjects: SceneObject[];
    readonly selectedObjects: SceneObject[];
    readonly unselectedObjects: SceneObject[];
    readonly highlightedObjects: SceneObject[];
    readonly unhighlightedObjects: SceneObject[];
    readonly pickableObjects: SceneObject[];
    readonly unpickableObjects: SceneObject[];
  }

  type SceneEvents = {
    "model-added": Model;
    "model-removed": Model;
    "object-added": SceneObject;
    "object-removed": SceneObject;
    "object-update": {
      object: SceneObject;
      property: keyof SceneStateObjectData | keyof Style | "geometry";
      value?: number | Array<number> | string;
      oldValue?: number | Array<number> | string;
    };
  };

  interface ModelData {
    readonly id?: string;
    readonly objects?: SceneObjectData[];
  }

  /**
   * A model is where objects live.
   */
  interface Model extends Transformable, Positionable, Typed<EntityType.model> {
    readonly id: string;
    dirty: boolean;
    readonly scene: Scene;
    addObject(objectData: SceneObjectData): SceneObject;
    removeObject(objectId: string): boolean;
    readonly objects: SceneObject[];
    readonly geometry: Positionable;
  }

  interface Translate {
    dx: number;
    dy: number;
  }

  interface SceneStateObjectData {
    visible: boolean;
    selected: boolean;
    highlighted: boolean;
    pickable: boolean;
  }

  interface SceneObjectData extends SceneStateObjectData, Style {
    readonly id?: string;
    readonly zIndex?: number;
    readonly pickingZIndex?: number;
    readonly geometry?: GeometryData | Array<LineData | Array<number>>;
  }

  interface SceneObject
    extends Style,
      Transformable,
      Positionable,
      Typed<EntityType.object> {
    readonly scene: Scene;
    readonly model: Model;
    readonly id: string;
    readonly zIndex: number;
    readonly pickingId: number;
    readonly pickingZIndex: number;
    readonly geometry: Geometry;
    flags: number;
  }

  type SceneObjectFactory = (
    objectData: SceneObjectData,
    model: Model
  ) => SceneObject;

  interface Drawable {
    draw(graphics: PIXIGraphics, lineSegments: number[]): void;
  }

  interface Geometry extends ReactiveEntity, Positionable {
    readonly shapes: Shape[];
    getShape(index?: number): Shape | undefined;
    addShape(shape: ShapeData | PointsData, index?: number): Shape;
    removeShape(index?: number): Shape | undefined;
  }

  type GeometryData = Array<number[] | ShapeData>;

  interface ShapeData {
    type?: ShapeType;
  }

  interface LineData extends ShapeData {
    type: ShapeType.line;
    points: PointsData;
    poly?: boolean;
  }

  interface ArcData extends ShapeData {
    type: ShapeType.arc;
    x: number;
    y: number;
    radius: number;
    startAngle?: number; // (0 is at the 3 o'clock position of the arc's circle)
    endAngle?: number;
    anticlockwise?: boolean;
    composite?: boolean;
  }

  interface Arc extends Shape {
    readonly type: ShapeType.arc;
    cache: ArcCache;
    x: number;
    y: number;
    radius: number;
    startAngle: number;
    endAngle: number;
    anticlockwise?: boolean;
  }

  interface ArcCache {
    aabb?: AABB;
  }

  enum ShapeType {
    line = "line",
    arc = "arc",
  }

  interface Shape
    extends ReactiveEntity,
      Drawable,
      Positionable,
      Transformable {
    readonly type: ShapeType;
    getAABB(): AABB;
    getCenter(): Point;
    cache: any;
  }

  type AABB = [minX: number, minY: number, maxX: number, maxY: number];

  interface Positionable {
    aabb: AABB;
    center: Point;
  }

  interface Transformable {
    move(position: Point): void;
    translate(dx: number, dy: number): void;
    /**
     * @param angle in radian, clockwise
     * @param origin
     */
    rotate(angle: number, origin?: Point): void;
    scale(ds: number, origin?: Point): void;
  }

  interface ReactiveEntity {
    onUpdate?(cfg: any): void;
  }

  interface Line extends Shape {
    readonly type: ShapeType.line;
    cache: LineCache;
    readonly poly: boolean;
    readonly points: Point[];
    getPoint(index?: number): Point | undefined;
    addPoint(point: Point, index?: number): Point;
    removePoint(index?: number): Point | undefined;
  }

  interface LineCache {
    aabb?: AABB;
    center?: Point;
  }

  type PointsData = number[];

  interface Style {
    // TEXTURE
    texture?: string;
    textureTint?: number;
    textureOpacity?: number;
    // LINE
    lineWidth?: number;
    lineColor?: number;
    lineOpacity?: number;
    lineDash?: number[];
    lineCap?: LINE_CAP;
    lineJoin?: LINE_JOIN;
  }

  interface TextureManager {
    textures: Map<string, NamedPIXITexture>;
    textureMatrix: TextureMatrix;
    load(
      name: string,
      src: string,
      width?: number,
      height?: number
    ): NamedPIXITexture;
    destroy(): void;
  }

  interface NamedPIXITexture extends PIXITexture {
    name: string;
  }

  interface TextureMatrix
    extends Readonly<
      EventHandler<{
        update: { rotate?: number; scale?: Point; translate?: Point };
      }>
    > {
    readonly matrix: PIXIMatrix;
    rotate(angle: number): void;
    scale(x: number, y?: number): void;
    translate(x: number, y: number): void;
  }

  interface SceneReport {
    addedObjects?: SceneObject[];
    removedObjects?: SceneObject[];
    visibleObjects?: SceneObject[];
    selectedObjects?: SceneObject[];
    highlightedObjects?: SceneObject[];
    pickableObjects?: SceneObject[];
    stylizedObjects?: SceneObject[];
    geometryObjects?: SceneObject[];
  }

  interface Renderer extends Readonly<EventHandler<RendererEvents>> {
    readonly viewer: Viewer;
    transform: PIXIMatrix;
    /**
     * @param { boolean } [force=false] The Renderer does not render each frame if nothing change. If `true`, it redraws the in memory geometries on the canvas.
     * @param { boolean } [clear=false] Clear the in memory geometries to redraw from scratch.
     */
    render(force?: boolean, clear?: boolean): void;
    transformUpdated: boolean;
    offset: Translate;
    readonly visibleContext: VisibleContext;
    readonly layers: Map<string, Layer>;
    resize(): void;
    pickColorAtCanvasPosition(
      position: Point,
      snap?: boolean
    ): number | undefined;
    makeIdDrawer(snap?: boolean): Drawer;
    makeStyleDrawer(style: Style, texture?: PIXITexture): Drawer;
  }

  type RendererEvents = {
    resize: { dx: number; dy: number };
    "pre-draw": undefined;
    "post-draw": undefined;
  };

  interface Drawer {
    displayObject: PIXIContainer;
    clear(): void;
    destroy(): void;
    draw(objects: SceneObject[]): void;
  }

  interface VisibleContext {
    readonly ctx: CanvasRenderingContext2D;
    resize(width: number, height: number): void;
    drawImage(canvas: HTMLCanvasElement, dx?: number, dy?: number): void;
    clear(): void;
  }

  interface Layer {
    name: string;
    visible: boolean;
    renderer: Renderer;
    _renderer: PIXIRenderer;
    view: HTMLCanvasElement;
    resize(width: number, height: number): void;
    stage: DirtyableContainer;
    objects: SceneObject[];
    addObjects(objects: SceneObject[]): void;
    updateObjects(objects: SceneObject[]): void;
    removeObjects(objects: SceneObject): void;
    render(
      forceDurty: boolean,
      forceRebuildStage: boolean,
      matrix: PIXIMatrix
    ): void;
    dirty: boolean;
    batches: Batch[];
    clear(): void;
    destroy(): void;
  }

  interface LayerOptions {
    select?: boolean;
    highlight?: boolean;
    picking?: boolean;
    showPicking?: boolean;
    snap?: boolean;
    showSnap?: boolean;
  }

  interface LayerHandlers {
    handleAddedObjects(): void;
    handleRemovedObjects(): void;
    handleUpdatedObjects(): void;
  }

  interface DirtyableContainer extends PIXIContainer {
    dirty: boolean;
  }

  interface Batch {
    dirty: boolean;
    addObjects(objects: SceneObject[]): void;
    removeObjects(objectIds: number[]): void;
    objects: Map<number, SceneObject>;
    destroy(): void;
  }

  interface StyleBatch extends Batch {
    modelId: string;
    key: string;
    zIndex: number;
  }
}
